import math
from typing import Optional
from aiogram import types


class InlineKeyboardPaginator:
    def __init__(
        self, 
        page_size: int, 
        previous_symbol: str = '<-',
        next_symbol: str = '->',
        back_symbol: str = 'back',
        separator: str = "_page:",
    ) -> None:
        self.page_size = page_size
        self.previous_symbol = previous_symbol
        self.next_symbol = next_symbol
        self.separator = separator
        self.back_symbol = back_symbol

    def paginate(
        self, 
        keyboard: types.InlineKeyboardMarkup, 
        page: int,
        all_count: int,
        data: str = "",
        back_data: Optional[str] = None,
    ) -> types.InlineKeyboardMarkup:
        if all_count < page:
            return keyboard

        final_page = math.ceil(all_count / self.page_size)
        pagination_row = []

        if page > 1:
            pagination_row.append(
               types.InlineKeyboardButton(
                    text=self.previous_symbol,
                    callback_data=f"{data}{self.separator}{(page - 1)}",
                )
            )
        if page < final_page:
            pagination_row.append(
               types.InlineKeyboardButton(
                    text=self.next_symbol,
                    callback_data=f"{data}{self.separator}{(page + 1)}",
                )
            )
        
        keyboard.inline_keyboard.append(pagination_row)

        if back_data is not None:
            keyboard.inline_keyboard.append(
                [
                   types.InlineKeyboardButton(
                        text=self.back_symbol,
                        callback_data=back_data,
                    )
                ]
            )
        
        return keyboard




from typing import Any, Coroutine, List, Union

from axsqlalchemy.repository import BaseRepository

from ..models import AbstractUserModel
from ..schemas import User as UserSchemas


class UserReposity(BaseRepository):
    DBModel = AbstractUserModel
    IModel = UserSchemas
    OModel = UserSchemas

    async def add(self, obj: IModel) -> Union[IModel, None]:
        res = await super().add(obj)

        if res:
            return self.OModel.from_orm(res)

    async def get(self, id: int) -> AbstractUserModel:
        return await super().get(*[id])



from typing import Callable, Optional, TypeVar, Iterable
from aiogram import types
from aiogram.utils.keyboard import InlineKeyboardBuilder


TObj = TypeVar('TObj')


def objs2inline(
    objs: Iterable[TObj], 
    getter: Callable[[TObj], types.InlineKeyboardButton]
) -> types.InlineKeyboardMarkup:
    builder = InlineKeyboardBuilder()
    
    for obj in objs:
        builder.add(getter(obj))

    builder.adjust(1)

    return builder.as_markup()


def ibutton(text: str, data: Optional[str] = None, url: Optional[str] = None) -> types.InlineKeyboardButton:
    return types.InlineKeyboardButton(
        text=text,
        callback_data=data,
        url=url,
    )




from functools import wraps
from typing import Any, Callable, TypeVar, Union, Protocol
from aiogram import F, Router, types

from axaiogram.utils.types import CallbackDataQuery, PaginatedMessage


class PaginatAbleHandler(Protocol):
    async def __call__(self, *args: Any, msg: types.Message, **kwds: Any) -> Any:
        raise NotImplementedError


def _extract_and_get_event_replaceable(
    args: list, 
    kwargs: dict
) -> Union[types.Message, CallbackDataQuery]:
    event_type = (types.Message, types.CallbackQuery)

    for i in range(len(args)):
        if type(args[i]) in event_type:
            return args.pop(i)
        
    for key, value in kwargs.items():
        if type(value) in event_type:
            return kwargs.pop(key)

    raise NotImplementedError


TPaginatAbleHandler = TypeVar('TPaginatAbleHandler', bound=Callable)  #, bound=PaginatAbleHandler)


def extract_paginate_page(
    separator: str = "_page:", 
    loading_msg: str = '...', 
    only_start_loading: bool = True,
    default_page=1,
):
    def inner(f: TPaginatAbleHandler) -> TPaginatAbleHandler:
        @wraps(f)
        async def wrapper(*args_, **kwargs_):
            args = list(args_)
            kwargs = dict(kwargs_)
            event= _extract_and_get_event_replaceable(args, kwargs)
            
            if not isinstance(event, types.CallbackQuery):
                msg = await event.answer(loading_msg)
                event_message = PaginatedMessage(user_message=event, cq=None, page=default_page, loading=msg)  # type: ignore
            else:
                new_data, *page = event.data.split(separator)
                page = ''.join(page)
                page = int(page) if page.strip().isdigit() else default_page
                if not only_start_loading:
                    msg = await event.message.edit_text(loading_msg)
                else:
                    msg = event.message
                event = event.copy(update={'data': new_data})
                event_message = PaginatedMessage(user_message=None, cq=event, page=page, loading=msg)  # type: ignore

            return await f(*args, msg=event_message, **kwargs)

        return wrapper  # type: ignore
    return inner


def register_as_paginated_handler(
    router: Router,
    handler: Callable,
    command_filter: str,
    *filters,
) -> Router:

    router.callback_query(F.text.startswith(command_filter), *filters)(handler)
    router.message(F.text.startswith(command_filter), *filters)(handler)

    return router




from abc import ABC, abstractclassmethod
from typing import Any, List, Type, Union

from pydantic import BaseModel

from .template_parser import TemplateParser


class TemplateGroup(ABC, BaseModel):
    base_path: str

    @property
    def template_parser(self) -> Type[TemplateParser]:
        raise NotImplementedError

    @abstractclassmethod
    def load(cls, langs: List[str]) -> "TemplateGroup":

        raise NotImplementedError

    def __getattribute__(self, __name: str) -> Union[TemplateParser, Any]:
        obj = super().__getattribute__(__name)

        if __name in self.__class__.__fields__.keys():
            return self.template_parser(base_formats=obj)

        return obj



from .on_the_fly_user_auth_middleware import OnTheFlyUserAuthMiddleware
from .use_handler_user_auth_middleware import UseHandlerUserAuthMiddleware

__all__ = [
    "OnTheFlyUserAuthMiddleware",
    "UseHandlerUserAuthMiddleware",
]



from typing import Union

from pydantic import BaseModel as _BaseModel


class BaseModel(_BaseModel):
    class Config:
        orm_mode = True


class User(BaseModel):
    id: int
    firstname: str
    username: str | None = None
    secondname: str | None = None
    is_active: bool = True


class UserLang(BaseModel):
    id: int
    lang_code: str | None = None


class UserWithEmail(User):
    email: str



from dataclasses import dataclass
from enum import Enum
from typing import Union

from aiogram import Bot, types
from aiogram.utils.keyboard import InlineKeyboardBuilder, ReplyKeyboardBuilder


class TemplateTextType:
    TEXT = "text"
    CAPTION = "caption"


class MessageTemplateTypeEnum:
    MESSAGE = "text"
    AUDIO = "audio"
    DOCUMENT = "document"
    PHOTO = "photo"
    STICKER = "sticker"
    VIDEO = "video"
    VIDEO_NOTE = "video_note"
    VOICE = "voice"
    CONTACT = "contact"


class TemplateKeyboardTypeEnum:
    INLINE = "inline"
    REPLY = "reply"


@dataclass
class TemplateMessage:
    type_: str
    id: str | None = None
    parse_mode: str | None = None

    text: Union[str, None] = None
    text_type: str = TemplateTextType.TEXT

    content_data: Union[dict, None] = None
    keyboard: Union[
        InlineKeyboardBuilder,
        ReplyKeyboardBuilder,
        types.ReplyKeyboardRemove,
        None,
    ] = None

    def to_message_data(self, **kwargs) -> dict:
        data = {
            self.text_type: self.text,
            "parse_mode": self.parse_mode,
            **(self.content_data or {}),
            **kwargs,
        }
        if type(self.keyboard) is not types.ReplyKeyboardRemove:
            data["reply_markup"] = self.keyboard.as_markup()
        else:
            data["reply_markup"] = self.keyboard

        return data



from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union
from aiogram import Bot, Router
from axabc.db import AbstractUOWFactory, AbstractUOW


TUOWFactory = TypeVar('TUOWFactory', bound=AbstractUOWFactory)


class BaseHandlersGroup(ABC):
    __router: Union[None, Router] = None
    uow: Optional[AbstractUOW] = None

    def __init__(self, bot: Bot, uowf: Optional[TUOWFactory] = None) -> None:
        self.bot = bot
        self.uowf = uowf
        self.router = proxy_router.get_group_observed_router(self)
        self.register_handlers(self.router)
    
    @abstractmethod
    def register_handlers(self, router: Router) -> Router:
        raise NotImplementedError

    @classmethod
    @property
    def router_collector(cls) -> Router:
        if not cls.__router:
            cls.__router = Router(name=cls.__name__)
        return cls.__router


def with_uow(f):
    @wraps(f)
    async def wrapper(self: BaseHandlersGroup, *args, **kwargs):
        if self.uowf is None:
            raise NotImplementedError('uowf is not found')

        async with self.uowf() as uow:  # type: ignore
            self.uow = uow
            return await f(self, *args, **kwargs)
    
    return wrapper 


OBSERVERS_NAME = [
    "message",
    "edited_message",
    "channel_post",
    "edited_channel_post",
    "inline_query",
    "chosen_inline_result",
    "callback_query",
    "shipping_query",
    "pre_checkout_query",
    "poll",
    "poll_answer",
    "my_chat_member",
    "chat_member",
    "chat_join_request",
    "error",
]

@dataclass
class FrozenHandler:
    observer_name: str

    args: Union[List[Any], tuple[Any]]
    kwargs: Dict[str, Any]
    fn: Any


class ProxyRouter(Router):
    _observers_name = OBSERVERS_NAME

    def __init__(self, *, name: str | None = None) -> None:
        self._cls_registrations: Dict[str, Dict[str, FrozenHandler]] = {}
        super().__init__(name=name)

    def proxy_observer(self, __name: str) -> Callable[[Any], Any]:
        def inner_wrapper(*args: Any, **kwargs: Dict[Any, Any]):
            def wrapper(fn: Callable[[Any], Any]) -> Any:
                clsname, methodname, *extra = fn.__qualname__.split('.')
                if not extra:
                    if clsname not in self._cls_registrations:
                        self._cls_registrations[clsname] = {}

                    self._cls_registrations[clsname][methodname] = FrozenHandler(
                        observer_name=__name,
                        args=args,
                        kwargs=kwargs,
                        fn=fn,
                    )

                return fn

            return wrapper

        return inner_wrapper

    def get_group_observed_router(self, group: BaseHandlersGroup) -> Router:
        cls = group.__class__
        clsname = cls.__name__
        router = Router(name=clsname)

        if clsname not in self._cls_registrations:
            return router

        for methodname, fh in self._cls_registrations[clsname].items():
            getattr(router, fh.observer_name)(*fh.args, **fh.kwargs)(getattr(group, methodname))

        return router


    def __getattribute__(self, __name: str) -> Any:
        if not __name.startswith('_') and __name in self._observers_name:
            return self.proxy_observer(__name)
        return super().__getattribute__(__name)


proxy_router = ProxyRouter()



from typing import Any, Coroutine, List, Union

from axsqlalchemy.repository import BaseRepository

from ..models import AbstractUserLanguageModel
from ..schemas import UserLang as UserLangSchemas


class UserLangReposity(BaseRepository):
    DBModel = AbstractUserLanguageModel
    IModel = UserLangSchemas
    OModel = UserLangSchemas

    async def add(self, obj: IModel) -> Any:
        return await super().add(obj)

    async def get(self, id: int) -> Any:
        return await super().get(id)



import inspect
from typing import Type, TypeVar

from templates.renderers.md.template_group import MarkdownTemplateGroup
from templates.template_group import TemplateGroup

TTemplateGroup = TypeVar("TTemplateGroup", bound=TemplateGroup)


class TemplateGroupFactory:
    def get_template_group(self, cls: Type[TTemplateGroup]) -> TemplateGroup:
        if cls.base_path.endswith(".md") and inspect.isclass(cls):

            class ImplementedTemplateGroup(cls, MarkdownTemplateGroup):
                ...

            return ImplementedTemplateGroup

        raise NotImplementedError



class HTMLTepmlateParser:
    def __init__(self, folder: str) -> None:
        self.folder = folder
        
    def parse(self, name: str, **kwargs):
        with open(f"{self.folder}/{name}", 'r') as f:
            content = f.read()
            return content.format(**kwargs)
        



from dataclasses import dataclass
from aiogram import types


class CallbackQuery(types.CallbackQuery):
    message: types.Message


class CallbackDataQuery(CallbackQuery):
    data: str


@dataclass
class PaginatedMessage:
    cq: CallbackDataQuery
    user_message: types.Message
    loading: types.Message
    page: int = 1




from typing import Any, Awaitable, Callable, Dict, Union

from aiogram import types
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import TelegramObject

from db.schemas import User as UserSchema

from ..templates import TemplateFactory


class TemplateMiddleWare(BaseMiddleware):
    def __init__(self, temp_factory: TemplateFactory, default_lang = 'en') -> None:
        self.temp_factory = temp_factory
        self.default_lang = default_lang
        super().__init__()

    async def __call__(
        self,
        handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: Dict[str, Any],
    ) -> Any:
        user: Union[UserSchema, None] = data.get("user")

        if user is None:
            raise NotImplementedError("TemplateMiddleware needs user data!")

        lang_code = user.lang_code or self.default_lang
        chat_id = None

        if type(event) is types.Message:
            chat_id = event.chat.id
        elif type(event) is types.CallbackQuery and event.message:
            chat_id = event.message.chat.id

        data["temp"] = self.temp_factory.get_template(lang_code, chat_id)

        return await handler(event, data)
